/************************************************************************
 *
 * Copyright 2009 J David Eisenberg All rights reserved.
 *
 * Uses ODF Toolkit which is Copyright 2008 Sun Microsystems, Inc.
 * All rights reserved.
 *
 * Use is subject to license terms.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at http://www.apache.org/licenses/LICENSE-2.0. You can also
 * obtain a copy of the License at http://odftoolkit.org/docs/license.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************************/
package org.odftoolkit.odfdom.examples;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import org.odftoolkit.odfdom.OdfElement;
import org.odftoolkit.odfdom.OdfFileDom;

import org.odftoolkit.odfdom.OdfNamespace;
import org.odftoolkit.odfdom.doc.office.OdfOfficeAutomaticStyles;
import org.odftoolkit.odfdom.doc.office.OdfOfficeStyles;


import org.odftoolkit.odfdom.doc.OdfTextDocument;
import org.odftoolkit.odfdom.doc.office.OdfOfficeText;
import org.odftoolkit.odfdom.doc.style.OdfStyle;
import org.odftoolkit.odfdom.doc.style.OdfStyleTextProperties;
import org.odftoolkit.odfdom.doc.text.OdfTextHeading;
import org.odftoolkit.odfdom.dom.attribute.text.TextStyleNameAttribute;
import org.odftoolkit.odfdom.dom.element.style.StyleTextPropertiesElement;
import org.odftoolkit.odfdom.dom.style.OdfStyleFamily;
import org.odftoolkit.odfdom.dom.style.props.OdfStyleProperty;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 * @author J David Eisenberg
 */
public class SearchByStyle {
    String inputFileName;
    String outputFileName;
    OdfTextDocument inputDocument;
    OdfTextDocument outputDocument;

	XPath inputXPath;	// use this for extracting information from input file

    OdfFileDom inputContentDom;	// the document object model for content.xml
    OdfFileDom inputStylesDom;	// the document object model for styles.xml

    OdfFileDom outputContentDom;	// the document object model for content.xml
    OdfFileDom outputStylesDom;	// the document object model for styles.xml

    // styles for input document
    OdfOfficeAutomaticStyles inputContentAutoStyles;
    OdfOfficeAutomaticStyles inputStylesAutoStyles;
    OdfOfficeStyles inputStylesOfficeStyles;

    // and output document
    OdfOfficeAutomaticStyles outputContentAutoStyles;
    OdfOfficeAutomaticStyles outputStylesAutoStyles;
    OdfOfficeStyles outputStylesOfficeStyles;

    OdfOfficeText inputOfficeText;
    OdfOfficeText outputOfficeText;

    HashMap<OdfStyleProperty,String> desiredProperties; // properties to look for
    ArrayList<String> desiredStyleNames; // styles that have those properties

    private void run(String[] args)
    {
         setupDocuments(args[0]);
        if (inputDocument != null && outputDocument != null)
        {
			inputXPath = XPathFactory.newInstance().newXPath();
            inputXPath.setNamespaceContext(new OdfNamespace());

            /* Set up the properties you want to look for */
            desiredProperties = new HashMap<OdfStyleProperty,String>();
            desiredProperties.put(StyleTextPropertiesElement.FontWeight,
                "bold");
            desiredProperties.put(StyleTextPropertiesElement.FontStyle,
                "italic");

            cleanOutOutputDocument();

            desiredStyleNames = new ArrayList<String>();
            gatherDesiredStyles(inputContentAutoStyles);
            gatherDesiredStyles(inputStylesAutoStyles);
            gatherDesiredStyles(inputStylesOfficeStyles);

            /* set up a custom style for the output */
            OdfStyle style = outputStylesOfficeStyles.newStyle("OutputHeading",
                OdfStyleFamily.Paragraph);
            style.setProperty(OdfStyleTextProperties.FontStyle,"italic");
            style.setProperty(OdfStyleTextProperties.FontWeight, "bold");
            style.setProperty(OdfStyleTextProperties.FontSize, "16pt");

            /* and insert a heading */
            outputOfficeText.appendChild(new
                OdfTextHeading(outputContentDom, "OutputHeading",
                "Featured Text"));

            seekAndOutputDesiredStyles();

            saveOutputDocument(args[1]);
        }
    }
    
    private void setupDocuments(String inputFilename)
	{
		try
		{
			outputDocument =
				OdfTextDocument.newTextDocument();
			outputOfficeText = outputDocument.getContentRoot();

			inputDocument = (OdfTextDocument)
                OdfTextDocument.loadDocument(inputFilename);
            inputOfficeText = inputDocument.getContentRoot();
           
            inputContentDom = inputDocument.getContentDom();
			inputStylesDom = inputDocument.getStylesDom();
			inputContentAutoStyles = inputContentDom.getOrCreateAutomaticStyles();
            inputStylesAutoStyles = inputStylesDom.getOrCreateAutomaticStyles();
			inputStylesOfficeStyles = inputDocument.getOrCreateDocumentStyles();

            outputContentDom = outputDocument.getContentDom();
			outputStylesDom = outputDocument.getStylesDom();
            outputContentAutoStyles = outputContentDom.getOrCreateAutomaticStyles();
            outputStylesAutoStyles = outputStylesDom.getOrCreateAutomaticStyles();
			outputStylesOfficeStyles = outputDocument.getOrCreateDocumentStyles();
		}
		catch (Exception e)
		{
			System.err.println("Unable to load input or create output.");
			System.err.println(e.getMessage());
            e.printStackTrace();
			outputDocument = null;
            inputDocument = null;
		}
	}
	/*
	 * The default document has some content in it already (in the case
	 * of a text document, a <text:p>.  Clean out all the old stuff.
	 */
	void cleanOutOutputDocument()
	{
		Node childNode;

		childNode = outputOfficeText.getFirstChild();
		while (childNode != null)
		{
			outputOfficeText.removeChild(childNode);
			childNode = outputOfficeText.getFirstChild();
		}
	}

    private void gatherDesiredStyles(Element styleElement)
    {
        NodeList styleNodes;
        OdfStyle style;
        Map<OdfStyleProperty,String> styleMap;
        String styleName;
        OdfStyleProperty propertyKey;
        String propertyValue;
        Iterator stepper;

		try {

            /* Use XPath to retrieve all the <style:style> nodes */
            styleNodes = (NodeList) inputXPath.evaluate("style:style",
                styleElement, XPathConstants.NODESET);
            for (int i = 0; i < styleNodes.getLength(); i++)
            {
                style = (OdfStyle) styleNodes.item(i);
                styleName = style.getStyleNameAttribute();

                /* Get all of this style's properties */
                styleMap = style.getStylePropertiesDeep();

                /*
                 * For each property we want to capture, see if
                 * this style has that property, set as we want it.
                 * If so, add it to the list of desired styles
                 */
                stepper = desiredProperties.keySet().iterator();
                while(stepper.hasNext()) {
                    propertyKey = (OdfStyleProperty) stepper.next();
                    propertyValue = desiredProperties.get(propertyKey);
                    if (styleMap.containsKey(propertyKey) &&
                    styleMap.get(propertyKey).equals(propertyValue))
                    {
                        desiredStyleNames.add(styleName);
                    }
                }
            }
		}
		catch (Exception e)
		{
			System.err.println("Unable to parse input file.");
			System.err.println(e.getMessage());
		}

    }

    /*
     * I'm taking the easy way out here, by letting XPath find
     * all the <text:p>, <text:span>, and <text:h> elements for me
     * rather than doing a recursive descent.
     *
     * For each of those elements, if its style name is one of the
     * ones I"m looking for, extract its text content and put that
     * content into the output document.
     */
    private void seekAndOutputDesiredStyles()
    {
        String styleName;
        String textContent;

        OdfElement element;
        try {
            inputXPath.reset();
            NodeList textNodes = (NodeList) inputXPath.evaluate(
                "//text:h|//text:p|//text:span",
            inputOfficeText, XPathConstants.NODESET);
            for (int i = 0; i < textNodes.getLength(); i++)
            {
                element = (OdfElement) textNodes.item(i);
                /*
                 * If you want to work at the convenience level,
                 * you have to go through this code to get the
                 * style name.
                 */
                /*
                if (element instanceof OdfTextParagraph)
                {
                    styleName = ((OdfTextParagraph) element).
                        getTextStyleNameAttribute();
                }
                else if (element instanceof OdfTextHeading)
                {
                    styleName = ((OdfTextHeading) element).
                        getTextStyleNameAttribute();
                }
                else
                {
                    styleName = ((OdfTextSpan) element).
                        getTextStyleNameAttribute();
                }
                */

                /*
                 * If you want to work at a low level, which is
                 * not encouraged, it gets a lot easier.
                 * As Don Marquis said, "It may not be moral,
                 * but it certainly would be expedient."
                 */
                styleName = element.getOdfAttributeValue(
                    TextStyleNameAttribute.ATTRIBUTE_NAME );
                
                if (desiredStyleNames.contains(styleName))
                {
                    textContent = collectAndNormalizeText(element);
                    outputDocument.newParagraph(textContent);
                }
            }
        }
        catch (XPathExpressionException e)
        {
            System.err.println("Could not evaluate expr in seekDesiredStyles");
			e.printStackTrace();
        }
        catch (Exception e)
        {
            System.err.println("Error while creating output document");
            e.printStackTrace();
        }

    }

    private String collectAndNormalizeText(Node node)
    {
        String result = "";
        String partial = "";
        String localName;
        /*
         * Use recursive descent to get all text elements.
         * If a node is entirely whitespace, convert to a single
         * blank. <text:tab>, <text:line-break>, and <text:s>
         * convert to a single blank.
         */
        node = node.getFirstChild();
        while (node != null)
        {
            if (node.getNodeType() == Node.TEXT_NODE)
            {
                partial = node.getNodeValue();
                if (partial.matches("^\\s+$"))
                {
                    partial = " ";
                }
                result += partial;
            }
            else if (node.getNodeType() == Node.ELEMENT_NODE)
            {
                localName = ((Element) node).getLocalName();
                if (localName.equals("s") || localName.equals("line-break")
                    || localName.equals("tab"))
                {
                    result += " ";
                }
                else
                {
                    result += collectAndNormalizeText(node);
                }
            }
            node = node.getNextSibling();
        }
        return result;
    }
    
    void saveOutputDocument(String outputFileName)
	{
		try
		{
			outputDocument.save(outputFileName);
		}
		catch (Exception e)
		{
			System.err.println("Unable to save document.");
			System.err.println(e.getMessage());
		}
	}

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        // TODO code application logic here
        SearchByStyle app = new SearchByStyle();
        app.run(args);
    }

}
